﻿module ParasitemiaCore.Analysis

open System
open System.Linq
open System.Drawing

open FSharp.Collections.ParallelSeq

open Emgu.CV
open Emgu.CV.Structure

open Logger

open Utils
open Morpho
open ImgTools
open Config
open Types

/// <summary>
/// Analyze the given image and detect reb blood cell (RBC) in it.
/// </summary>
/// <param name="img">The image</param>
/// <param name="name">The name, used during logging</param>
/// <param name="config">The configuration, must not be shared with another analysis</param>
/// <param name="reportProgress">An optional function to report progress and/or cancel the process.
///     The first call returning 'false' will cancel the analysis.
///     The 'int' parameter correspond to the progression from 0 to 100</param>
/// <returns>A list of detected cells or nothing if the process has been cancelled</returns>
let doAnalysis (img : Image<Bgr, byte>) (name : string) (config : Config) (reportProgress : (int -> bool) option) : Cell list option =

    // To report the progress of this function from 0 to 100.
    // Return 'None' if the process must be aborted.
    let reportWithVal (percent : int) (value : 'a) : 'a option =
        match reportProgress with
        | Some f -> if f percent then Some value else None
        | _ -> Some value

    let report (percent : int) : unit option =
        reportWithVal percent ()

    let inline buildLogWithName (text : string) = sprintf "№ %s: %s" name text
    let logWithName mess = Log.Info "%s" (buildLogWithName mess)
    let inline logTimeWithName (text : string) (f : unit -> 'a option) : 'a option = Log.LogWithTime Severity.INFO f "%s" (buildLogWithName text)

    // Monadic construction to be able to abort the progress when running.
    maybe {
        do! report 0

        logWithName "Starting analysis ..."

        use img_float = img.Convert<Bgr, float32> ()

        use img_RBC = img_float.[1] // Green.
        use img_RBC_filtered = gaussianFilter img_RBC config.LPFStandardDeviationRBC

        use img_parasites = img_float.[2] // Red.
        use img_parasites_filtered = gaussianFilter img_parasites config.LPFStandardDeviationParasite

        logWithName (sprintf "Nominal erythrocyte diameter: %O" config.RBCRadiusByResolution)

        let initialAreaOpening = int <| config.RBCRadiusByResolution.Area * config.Parameters.ratioAreaPaleCenter * 1.1f // We do an area opening a little larger to avoid to do a second one in the case the radius found is near the initial one.
        do! logTimeWithName "First area opening" (fun () -> areaOpenF img_RBC_filtered initialAreaOpening; report 10)

        let range =
            let delta = config.Parameters.granulometryRange * config.RBCRadiusByResolution.Pixel
            int <| config.RBCRadiusByResolution.Pixel - delta, int <| config.RBCRadiusByResolution.Pixel + delta

        let! radius = logTimeWithName "Granulometry (area)" (fun () -> reportWithVal 10 (Granulometry.findRadiusByAreaClosing img_RBC_filtered range |> float32))
        //let! radius = logTimeWithName "Granulometry (morpho)" (fun () -> reportWithVal 10 (Granulometry.findRadiusByClosing img_RBC_filtered range 1. true |> float32))
        config.SetRBCRadius <| radius

        logWithName (sprintf "Found erythrocyte diameter: %O" config.RBCRadius)

        do! report 20

        do!
            let secondAreaOpening = int <| config.RBCRadius.Area * config.Parameters.ratioAreaPaleCenter
            if secondAreaOpening > initialAreaOpening then
                logTimeWithName "Second area opening" (fun () -> areaOpenF img_RBC_filtered secondAreaOpening; report 30)
            else
                report 30

        // Remove pale centers from the parasites image.
        areaOpenF img_parasites_filtered (int <| config.RBCRadiusByResolution.Area * config.Parameters.ratioAreaPaleCenter)

        // Removing parasites.
        areaCloseF img_RBC_filtered (roundInt <| Const.PI * config.RBCRadius.ParasiteRadius ** 2.f)

        let! parasites, imgWhitoutParasite, imgWithoutNucleus =
            logTimeWithName "Parasites segmentation" (fun () -> reportWithVal 40 (ParasitesMarker.find img_parasites_filtered config))

        let! edges, xGradient, yGradient =
            logTimeWithName "Finding edges" (
                fun () ->
                    let edges, xGradient, yGradient = Edges.find img_RBC_filtered
                    removeArea edges (config.RBCRadius.Pixel ** 2.f / 50.f |> int)
                    reportWithVal 50 (edges, xGradient, yGradient)
            )

        let! matchingEllipses = logTimeWithName "Finding ellipses" (fun () -> reportWithVal 60 (Ellipse.find edges xGradient yGradient config))

        let! prunedEllipses = logTimeWithName "Ellipses pruning" (fun () -> reportWithVal 80 (matchingEllipses.PrunedEllipses))

        let! cells = logTimeWithName "Classifier" (fun () -> reportWithVal 100 (Classifier.findCells prunedEllipses parasites img.Width img.Height config))

        logWithName "Analysis finished"

        do
            // Output pictures if debug flag is set.
            match config.Debug with
            | DebugOn output ->
                let dirPath = System.IO.Path.Combine (output, name)
                System.IO.Directory.CreateDirectory dirPath |> ignore

                let buildFileName postfix = System.IO.Path.Combine (dirPath, name + postfix)

                IO.saveMat (edges * 255.0) (buildFileName " - edges.png")

                IO.saveImg parasites.darkStain (buildFileName " - parasites - dark stain.png")
                IO.saveImg parasites.parasite (buildFileName " - parasites - stain.png")
                IO.saveImg parasites.nucleus (buildFileName " - parasites - infection.png")

                let imgAllEllipses = img_RBC_filtered.Copy ()
                Drawing.drawEllipses imgAllEllipses matchingEllipses.Ellipses (Gray 200.0) 0.04
                IO.saveImg imgAllEllipses (buildFileName " - ellipses - all.png")

                let imgEllipses = img_RBC_filtered.Convert<Bgr, byte> ()
                Drawing.drawEllipses imgEllipses prunedEllipses (Bgr (0.0, 240.0, 240.0)) 1.0
                IO.saveImg imgEllipses (buildFileName " - ellipses.png")

                let imgCells = img.Copy ()
                Drawing.drawCells imgCells false cells
                IO.saveImg imgCells (buildFileName " - cells.png")

                let imgCells' = img.Copy ()
                Drawing.drawCells imgCells' true cells
                IO.saveImg imgCells' (buildFileName " - cells - full.png")

                (* let filteredRBCMaxima = gaussianFilter img_RBC config.LPFStandardDeviationRBC
                for m in findMaxima filteredRBCMaxima do
                    Drawing.drawPoints filteredRBCMaxima m 255.f
                IO.saveImg filteredRBCMaxima (buildFileName " - filtered - maxima.png") *)

                IO.saveImg imgWhitoutParasite (buildFileName " - filtered closed stain.png")
                IO.saveImg imgWithoutNucleus (buildFileName " - filtered closed infection.png")

                IO.saveImg img_RBC_filtered (buildFileName " - source - RBC.png")
                IO.saveImg img_parasites_filtered (buildFileName " - source - parasites.png")

                IO.saveImg img_float.[2] (buildFileName " - source - red.png")
                IO.saveImg img_float.[1] (buildFileName " - source - green.png")
                IO.saveImg img_float.[0] (buildFileName " - source - blue.png")
            | _ -> ()

        return cells
    }

/// <summary>
/// Do multiple analyses on the same time. The number of concurrent process depends if the number of the core.
/// </summary>
/// <param name="imgs">The images: (name * configuration * image)</param>
/// <param name="reportProgress">An optional function to report progress and/or cancel the process.
///     The first call returning 'false' will cancel the analysis.
///     The 'int' parameter correspond to the progression from 0 to 100</param>
/// <returns>'None' if the process has been cancelled or the list of result as (name * cells), 'name' corresponds to the given name<returns>
let doMultipleAnalysis (imgs : (string * Config * Image<Bgr, byte>) list) (reportProgress : (int -> bool) option) : (string * Cell list) list option =
    let report (percent : int) : bool =
        match reportProgress with
        | Some f -> f percent
        | _ -> true

    let progressPerAnalysis = System.Collections.Concurrent.ConcurrentDictionary<string, int> ()
    let nbImgs = List.length imgs

    let reportProgressImg (id : string) (progress : int) =
        progressPerAnalysis.AddOrUpdate (id, progress, (fun _ _ -> progress)) |> ignore
        report (progressPerAnalysis.Values.Sum () / nbImgs)

    let n = Environment.ProcessorCount

    let results =
        imgs
        |> PSeq.choose (
            fun (id, config, img) ->
                try
                    match doAnalysis img id config (Some (fun p -> reportProgressImg id p)) with
                    | Some result -> Some (id, result)
                    | None -> None
                with
                | ex ->
                    Log.Error "Analysis %s failed: %O" id ex
                    None
        )
        |> PSeq.withDegreeOfParallelism n
        |> PSeq.toList

    // If one of the analyses has been aborted we return 'None'.
    if List.length results <> List.length imgs then
        None
    else
        Some results

